package ru.dbotthepony.mc.otm.network.syncher

import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap
import it.unimi.dsi.fastutil.ints.Int2ReferenceLinkedOpenHashMap
import it.unimi.dsi.fastutil.ints.Int2ReferenceMap
import it.unimi.dsi.fastutil.ints.Int2ReferenceOpenHashMap
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.ints.IntIterator
import it.unimi.dsi.fastutil.ints.IntSet
import it.unimi.dsi.fastutil.objects.ObjectIterator
import it.unimi.dsi.fastutil.objects.ObjectSet
import it.unimi.dsi.fastutil.objects.Reference2ObjectLinkedOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceCollection
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceSet
import net.minecraft.network.RegistryFriendlyByteBuf
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.mc.otm.util.IDAllocator
import java.io.Closeable
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.atomic.AtomicBoolean

/**
 * Constructs a new [DynamicSynchableGroup] where [T] is itself an [ISynchable]
 */
fun <T : ISynchable> DynamicSynchableGroup(reader: RegistryFriendlyByteBuf.() -> T, writer: T.(RegistryFriendlyByteBuf) -> Unit): DynamicSynchableGroup<T> {
	return DynamicSynchableGroup(reader, { this }, writer)
}

/**
 * Syncher group/set, which deals with synchables of only one type, and which are created and removed
 * on remote (e.g. server adding and removing synchables at will), which makes it distinct from
 * [SynchableGroup], in which attached synchables are created/removed manually on both sides.
 */
class DynamicSynchableGroup<T : Any>(
	/**
	 * Constructs new [T] instance locally, when remote created one.
	 * Data written by [writer] must be read here, if there is any.
	 */
	private val reader: RegistryFriendlyByteBuf.() -> T,

	/**
	 * Gets an [ISynchable] out of value [T] for synching state
	 */
	private val synchable: T.() -> ISynchable,

	/**
	 * Allows to write additional data to network stream during
	 * first-time networking of [T] to remote
	 */
	private val writer: T.(RegistryFriendlyByteBuf) -> Unit = {},
) : ISynchable, ReferenceSet<T> {
	private inner class RemoteState(val listener: Runnable) : IRemoteState {
		private inner class RemoteSlot(val slot: Slot<T>, fromConstructor: Boolean) : Runnable, Closeable {
			val remoteState = synchable(slot.value).createRemoteState(this)
			val isDirty = AtomicBoolean(true)
			var isRemoved = false
				private set

			init {
				if (!fromConstructor)
					removals.remove(slot.id)

				firstTime.add(this)
				dirty.add(this)

				if (firstTime.size != 1 && !fromConstructor) {
					listener.run()
				}
			}

			override fun run() {
				if (isDirty.compareAndSet(false, true)) {
					dirty.add(this)
					listener.run()
				}
			}

			fun markClean() {
				isDirty.set(false)
			}

			override fun close() {
				isRemoved = true
				isDirty.set(true)
				remoteState.close()
			}
		}

		private val remotes = HashMap<Slot<T>, RemoteSlot>()
		private val dirty = ConcurrentLinkedQueue<RemoteSlot>()
		private val firstTime = ArrayList<RemoteSlot>()
		private val removals = IntArraySet()
		private var sendClear = false

		init {
			value2slot.values.forEach {
				remotes[it] = RemoteSlot(it, true)
			}

			remoteStates.add(this)
		}

		fun add(element: Slot<T>) {
			remotes[element] = RemoteSlot(element, false)
			listener.run()
		}

		fun remove(element: Slot<T>) {
			val removed = remotes.remove(element)
			checkNotNull(removed)
			check(removed.slot == element)
			removed.close()
			removals.add(element.id)
			listener.run()
		}

		override fun write(stream: RegistryFriendlyByteBuf) {
			if (sendClear) {
				sendClear = false
				stream.writeByte(CLEAR)
			}

			run {
				val itr = removals.iterator()

				while (itr.hasNext()) {
					stream.writeByte(REMOVE_ENTRY)
					stream.writeVarInt(itr.nextInt())
				}

				removals.clear()
			}

			firstTime.forEach {
				stream.writeByte(ADD_ENTRY)
				stream.writeVarInt(it.slot.id)
				writer(it.slot.value, stream)
			}

			firstTime.clear()

			val seen = ReferenceOpenHashSet<RemoteSlot>()
			val slots = ArrayList<RemoteSlot>()
			var next = dirty.poll()

			while (next != null) {
				if (!next.isRemoved) {
					val status = seen.add(next)
					next.markClean()

					if (status) {
						slots.add(next)
					}
				}

				next = dirty.poll()
			}

			slots.forEach {
				stream.writeByte(SYNC_ENTRY)
				stream.writeVarInt(it.slot.id)
				it.remoteState.write(stream)
			}

			stream.writeByte(END)
		}

		override fun invalidate() {
			firstTime.clear()
			dirty.clear()
			removals.clear() // is it necessary?

			remotes.values.forEach {
				firstTime.add(it)
				dirty.add(it)
			}
		}

		override fun close() {
			remotes.values.forEach { it.close() }

			firstTime.clear()
			dirty.clear()
			removals.clear()
			remotes.clear()

			remoteStates.remove(this)
		}

		fun clear() {
			sendClear = true
			firstTime.clear()
			dirty.clear()
			removals.clear()
			remotes.values.forEach { it.close() }
			remotes.clear()
		}
	}

	private class Slot<T : Any>(val value: T, val id: Int)

	private var defaultReturnValue: T? = null
	private val remoteStates = ArrayList<RemoteState>()
	private val value2slot = Reference2ObjectLinkedOpenHashMap<T, Slot<T>>()
	private val id2slot = Int2ReferenceLinkedOpenHashMap<Slot<T>>()
	private val idAllocator = IDAllocator()

	override val hasRemotes: Boolean
		get() = remoteStates.isNotEmpty()

	override var isRemote: Boolean = false
		private set

	override fun read(stream: RegistryFriendlyByteBuf) {
		isRemote = true
		var action = stream.readByte().toInt()

		while (action != END) {
			when (action) {
				REMOVE_ENTRY -> {
					val id = stream.readVarInt()
					val slot = checkNotNull(id2slot.remove(id)) { "No such slot with ID: $id" }
					check(value2slot.remove(slot.value) == slot)
					remoteStates.forEach { it.remove(slot) }
				}

				ADD_ENTRY -> {
					val id = stream.readVarInt()

					// it is expected that during very frequent updates entry gets removed and new entry takes up their ID,
					// so we allow implicit removal if ID is allocated by some other entry before old entry could have a chance
					// to broadcast that it have been removed

					if (id2slot.containsKey(id)) {
						val slot = id2slot.remove(id)!!
						check(value2slot.remove(slot.value) == slot)
						remoteStates.forEach { it.remove(slot) }
					}

					val value = reader(stream)
					val slot = Slot(value, id)
					value2slot[value] = slot
					id2slot[id] = slot
				}

				SYNC_ENTRY -> {
					val id = stream.readVarInt()
					val slot = checkNotNull(id2slot.get(id)) { "No such slot with ID: $id" }
					synchable(slot.value).read(stream)
				}

				CLEAR -> {
					value2slot.clear()
					id2slot.clear()
				}

				else -> throw IllegalArgumentException("Unknown network action: $action")
			}

			action = stream.readByte().toInt()
		}
	}

	override fun createRemoteState(listener: Runnable): IRemoteState {
		return RemoteState(listener)
	}

	override fun add(element: T): Boolean {
		if (element in value2slot) {
			return false
		}

		val slot = Slot(element, idAllocator.allocate())
		value2slot[element] = slot
		id2slot[slot.id] = slot
		remoteStates.forEach { it.add(slot) }
		return true
	}

	override val size: Int
		get() = value2slot.size

	override fun addAll(elements: Collection<T>): Boolean {
		var any = false
		elements.forEach { any = add(it) || any }
		return any
	}

	override fun clear() {
		if (value2slot.isNotEmpty()) {
			value2slot.clear()
			id2slot.clear()
			idAllocator.reset()

			remoteStates.toTypedArray().forEach { it.clear() }
		}
	}

	override fun contains(element: T): Boolean {
		return element in value2slot
	}

	override fun containsAll(elements: Collection<T>): Boolean {
		return value2slot.keys.containsAll(elements)
	}

	override fun isEmpty(): Boolean {
		return value2slot.isEmpty()
	}

	override fun iterator(): ObjectIterator<T> {
		return object : ObjectIterator<T> {
			private val parent = value2slot.values.iterator()
			private var last: KOptional<Slot<T>> = KOptional()

			override fun hasNext(): Boolean {
				return parent.hasNext()
			}

			override fun next(): T {
				val slot = parent.next()
				last = KOptional(slot)
				return slot.value
			}

			override fun remove() {
				parent.remove()
				val slot = last.value
				checkNotNull(id2slot.remove(slot.id))
				remoteStates.forEach { it.remove(slot) }
				last = KOptional()
			}
		}
	}

	private fun removeBySlot(slot: Slot<T>) {
		value2slot.remove(slot.value)
		id2slot.remove(slot.id)
		remoteStates.forEach { it.remove(slot) }
		idAllocator.release(slot.id)
	}

	override fun remove(element: T): Boolean {
		removeBySlot(value2slot.remove(element) ?: return false)
		return true
	}

	override fun removeAll(elements: Collection<T>): Boolean {
		var any = false
		elements.forEach { any = remove(it) || any }
		return any
	}

	override fun retainAll(elements: Collection<T>): Boolean {
		val itr = iterator()
		var any = false

		for (element in itr) {
			if (element !in elements) {
				itr.remove()
				any = true
			}
		}

		return any
	}

	/**
	 * Representation of this synchable group as an [Int2ReferenceMap]. Changes made to [DynamicSynchableGroup] are reflected in this map
	 * and vice versa. Entries can be put at arbitrary keys, but negative keys should be avoided, since they are not efficiently
	 * networked, and hence will create a lot of network traffic if synchables with negative keys get frequently updated.
	 */
	val asMap: Int2ReferenceMap<T> by lazy {
		object : Int2ReferenceMap<T> {
			override fun isEmpty(): Boolean {
				return this@DynamicSynchableGroup.isEmpty()
			}

			override fun remove(key: Int): T? {
				val slot = id2slot.remove(key) ?: return null
				removeBySlot(slot)
				return slot.value
			}

			override fun getOrDefault(key: Int, defaultValue: T): T {
				return super.getOrDefault(key, defaultValue)
			}

			override val size: Int
				get() = this@DynamicSynchableGroup.size

			override fun put(key: Int, value: T): T? {
				val slot = Slot(value, key)
				val oldSlot = id2slot.put(key, slot)

				if (oldSlot?.value === value)
					return value

				if (oldSlot != null) {
					remoteStates.forEach { it.remove(oldSlot) }
					value2slot.remove(oldSlot.value)
				} else {
					idAllocator.allocate(key)
				}

				value2slot[value] = slot
				remoteStates.forEach { it.add(slot) }

				return oldSlot?.value
			}

			override fun get(key: Int): T? {
				return id2slot.get(key)?.value ?: defaultReturnValue
			}

			override fun containsKey(key: Int): Boolean {
				return id2slot.containsKey(key)
			}

			override fun defaultReturnValue(rv: T?) {
				defaultReturnValue = rv
			}

			override fun defaultReturnValue(): T? {
				return defaultReturnValue
			}

			override fun putAll(from: Map<out Int, T>) {
				from.forEach { (k, v) ->
					put(k, v)
				}
			}

			override fun containsValue(value: T): Boolean {
				return value2slot.containsKey(value)
			}

			private val entrySet: ObjectSet<Int2ReferenceMap.Entry<T>> by lazy(LazyThreadSafetyMode.PUBLICATION) {
				object : ObjectSet<Int2ReferenceMap.Entry<T>> {
					inner class Entry(private val ikey: Int) : Int2ReferenceMap.Entry<T> {
						override val value: T
							get() = id2slot[ikey]?.value ?: throw NoSuchElementException()

						override fun setValue(newValue: T): T? {
							return put(ikey, newValue)
						}

						override fun getIntKey(): Int {
							return ikey
						}

						override fun equals(other: Any?): Boolean {
							return this === other || other is Int2ReferenceMap.Entry<*> && other.intKey == ikey && other.value === id2slot[ikey]?.value
						}

						override fun hashCode(): Int {
							return ikey * 31 + (id2slot[ikey] as Slot<T>?).hashCode()
						}

						override fun toString(): String {
							return "Map.Entry[$ikey, ${id2slot[ikey]?.value}]"
						}
					}

					override val size: Int
						get() = id2slot.size

					override fun add(element: Int2ReferenceMap.Entry<T>): Boolean {
						return put(element.intKey, element.value) != null
					}

					override fun addAll(elements: Collection<Int2ReferenceMap.Entry<T>>): Boolean {
						var any = false
						elements.forEach { any = add(it) || any }
						return any
					}

					override fun clear() {
						this@DynamicSynchableGroup.clear()
					}

					override fun iterator(): ObjectIterator<Int2ReferenceMap.Entry<T>> {
						return object : ObjectIterator<Int2ReferenceMap.Entry<T>> {
							private val parent = id2slot.int2ReferenceEntrySet().iterator()
							private var last: Slot<T>? = null

							override fun hasNext(): Boolean {
								return parent.hasNext()
							}

							override fun remove() {
								if (last == null)
									throw NoSuchElementException()

								parent.remove()
								removeBySlot(last!!)
								last = null
							}

							override fun next(): Int2ReferenceMap.Entry<T> {
								val value = parent.next()
								last = value.value
								return Entry(value.intKey)
							}
						}
					}

					override fun remove(element: Int2ReferenceMap.Entry<T>): Boolean {
						val slot = id2slot.get(element.intKey)

						if (slot?.value === element.value) {
							this@DynamicSynchableGroup.removeBySlot(slot)
							return true
						}

						return false
					}

					override fun removeAll(elements: Collection<Int2ReferenceMap.Entry<T>>): Boolean {
						var any = false
						elements.forEach { any = remove(it) || any }
						return any
					}

					override fun retainAll(elements: Collection<Int2ReferenceMap.Entry<T>>): Boolean {
						if (elements.isEmpty()) {
							val isNotEmpty = id2slot.isNotEmpty()
							this@DynamicSynchableGroup.clear()
							return isNotEmpty
						}

						var any = false
						val itr = id2slot.int2ReferenceEntrySet().iterator()

						for (entry in itr) {
							if (elements.none { it.intKey == entry.intKey && it.value === entry.value.value }) {
								any = true
								itr.remove()
								removeBySlot(entry.value)
							}
						}

						return any
					}

					override fun contains(element: Int2ReferenceMap.Entry<T>): Boolean {
						return id2slot.get(element.intKey)?.value === element.value
					}

					override fun containsAll(elements: Collection<Int2ReferenceMap.Entry<T>>): Boolean {
						return elements.all { contains(it) }
					}

					override fun isEmpty(): Boolean {
						return id2slot.isEmpty()
					}
				}
			}

			override fun int2ReferenceEntrySet(): ObjectSet<Int2ReferenceMap.Entry<T>> {
				return entrySet
			}

			override val keys: IntSet by lazy(LazyThreadSafetyMode.PUBLICATION) {
				object : IntSet {
					override fun add(key: Int): Boolean {
						throw UnsupportedOperationException()
					}

					override fun addAll(c: IntCollection): Boolean {
						throw UnsupportedOperationException()
					}

					override fun addAll(elements: Collection<Int>): Boolean {
						throw UnsupportedOperationException()
					}

					override fun clear() {
						this@DynamicSynchableGroup.clear()
					}

					override fun iterator(): IntIterator {
						return object : IntIterator {
							private val parent = id2slot.keys.iterator()
							private var last = -1
							private var hasLast = false

							override fun hasNext(): Boolean {
								return parent.hasNext()
							}

							override fun remove() {
								if (!hasLast)
									throw NoSuchElementException()

								hasLast = false
								val slot = id2slot.get(last)
								parent.remove()
								removeBySlot(slot)
							}

							override fun nextInt(): Int {
								last = parent.nextInt()
								hasLast = true
								return last
							}
						}
					}

					override fun remove(k: Int): Boolean {
						removeBySlot(id2slot.get(k) ?: return false)
						return true
					}

					override fun removeAll(c: IntCollection): Boolean {
						val itr = c.iterator()
						var any = false

						while (itr.hasNext()) {
							any = remove(itr.nextInt()) || any
						}

						return any
					}

					override fun removeAll(elements: Collection<Int>): Boolean {
						val itr = elements.iterator()
						var any = false

						while (itr.hasNext()) {
							any = remove(itr.next()) || any
						}

						return any
					}

					override fun retainAll(c: IntCollection): Boolean {
						val itr = id2slot.values.iterator()
						var any = false

						for (slot in itr) {
							if (!c.contains(slot.id)) {
								itr.remove()
								check(value2slot.remove(slot.value) == slot)
								remoteStates.forEach { it.remove(slot) }
								any = true
							}
						}

						return any
					}

					override fun retainAll(elements: Collection<Int>): Boolean {
						val itr = id2slot.values.iterator()
						var any = false

						for (slot in itr) {
							if (!elements.contains(slot.id)) {
								itr.remove()
								check(value2slot.remove(slot.value) == slot)
								remoteStates.forEach { it.remove(slot) }
								any = true
							}
						}

						return any
					}

					override fun contains(key: Int): Boolean {
						return id2slot.containsKey(key)
					}

					override fun containsAll(c: IntCollection): Boolean {
						return id2slot.keys.containsAll(c)
					}

					override fun containsAll(elements: Collection<Int>): Boolean {
						return id2slot.keys.containsAll(elements)
					}

					override fun isEmpty(): Boolean {
						return id2slot.isEmpty()
					}

					override fun toArray(a: IntArray?): IntArray {
						return id2slot.keys.toArray(a)
					}

					override fun toIntArray(): IntArray {
						return id2slot.keys.toIntArray()
					}

					override val size: Int
						get() = id2slot.size
				}
			}

			override val values: ReferenceCollection<T>
				get() = this@DynamicSynchableGroup
		}
	}

	companion object {
		private const val END = 0
		private const val ADD_ENTRY = 1
		private const val SYNC_ENTRY = 2
		private const val REMOVE_ENTRY = 3
		private const val CLEAR = 4
	}
}
